/* ==========================================================[ CommandSet.cc ]=
 * CommandSet.cc - command callback functions
 * Copyright (C) 2002 Hubert Chan <hubert@uhoreg.ca>
 *
 * ----------------------------------------------------------------( License )-
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * ------------------------------------------------------------( Description )-
 *
 * This file maps command names to the functions implementing the commands.
 * Commands are attached to modes of the player, or can be accessible in all
 * modes.
 *
 * ------------------------------------------------------------------------- */

#include "CommandSet.hh"

// static members
std::vector<std::map<std::string,CommandSet::callback_t> > CommandSet::callbacks(NUM_MODES);
bool CommandSet::is_initialized;
std::map<std::string,CommandSet::callback_t> &CommandSet::all_modes = CommandSet::callbacks[MODE_ALL];

void CommandSet::modeChangedCallback()
{
  curr_mode = &callbacks[sessiondata.getMode()];
  switch (sessiondata.getMode())
  {
  case MODE_PLAYER:
    selectedSong = -1;
    sessiondata.display->setSelectedSong(-1);
    playlistMode = PLMODE_PLAY;
    sessiondata.display->setPlaylistMode(PLMODE_PLAY);

    selectedPref = 0;
    sessiondata.display->setPreference(0);
    break;
  case MODE_PLAYLIST:
    while (sessiondata.selector->backout())
      selectedItemHistory.pop();
    sessiondata.display->updateSelectionList();
    sessiondata.display->setSelection(selectedItemHistory.top());
    break;
  default:
    break;
  }
}

void CommandSet::play(const std::string &args)
{
  if (sessiondata.coreplayer->IsPlaying())
  { // if paused, resume
    if (sessiondata.coreplayer->GetSpeed() == 0.0)
    {
      sessiondata.display->playerUnpaused();
      sessiondata.coreplayer->SetSpeed(oldspeed);
    }
  }
  else
  {
    sessiondata.playlist->UnPause();
    sessiondata.playlist->Play(0);
  }
}

void CommandSet::stop(const std::string &args)
{
  sessiondata.playlist->Stop();
  if (sessiondata.playlist->Length())
    sessiondata.playlist->SetCurrent(1);
}

void CommandSet::pause(const std::string &args)
{
  sessiondata.display->playerPaused();
  oldspeed = sessiondata.coreplayer->GetSpeed();
  sessiondata.coreplayer->SetSpeed(0.0);
}

void CommandSet::playpause(const std::string &args)
{
  if (sessiondata.coreplayer->GetSpeed() == 0.0)
  {
    sessiondata.display->playerUnpaused();
    sessiondata.coreplayer->SetSpeed(oldspeed);
  }
  else
  {
    sessiondata.display->playerPaused();
    oldspeed = sessiondata.coreplayer->GetSpeed();
    sessiondata.coreplayer->SetSpeed(0.0);
  }
}

void CommandSet::next(const std::string &args)
{
  if (args.size() > 0U)
  {
    int arg_int = atoi(args.c_str());
    for (int i = 0; i < arg_int; i++)
      sessiondata.playlist->Next();
  }
  else
    sessiondata.playlist->Next();
}

void CommandSet::prev(const std::string &args)
{
  if (args.size() > 0U)
  {
    int arg_int = atoi(args.c_str());
    for (int i = 0; i < arg_int; i++)
      sessiondata.playlist->Prev();
  }
  else
    sessiondata.playlist->Prev();
}

void CommandSet::fwd(const std::string &args)
{
  int arg_int;
  if (args.size() == 0U)
    arg_int = 1;
  else
    arg_int = atoi(args.c_str());
  int cur_pos = (sessiondata.coreplayer->GetCurrentTime() / 100) + arg_int;
  if (cur_pos < 0)
    cur_pos = 0;
  cur_pos *= sessiondata.coreplayer->GetSampleRate();
  cur_pos /= sessiondata.coreplayer->GetFrameSize();
  cur_pos *=  sessiondata.coreplayer->GetChannels();
  cur_pos *= 2; // 16-bit = 2 * 1 byte
  sessiondata.coreplayer->Seek(cur_pos);
}

void CommandSet::bwd(const std::string &args)
{
  int arg_int;
  if (args.size() == 0U)
    arg_int = 1;
  else
    arg_int = atoi(args.c_str());
  int cur_pos = (sessiondata.coreplayer->GetCurrentTime() / 100) - arg_int;
  if (cur_pos < 0)
    cur_pos = 0;
  cur_pos *= sessiondata.coreplayer->GetSampleRate();
  cur_pos /= sessiondata.coreplayer->GetFrameSize();
  cur_pos *=  sessiondata.coreplayer->GetChannels();
  cur_pos *= 2; // 16-bit = 2 * 1 byte
  sessiondata.coreplayer->Seek(cur_pos);
}

void CommandSet::vol_up(const std::string &args)
{
  int volume, arg_int;
  if (args.size() == 0U)
    arg_int = 5;
  else
    arg_int = atoi(args.c_str());
  volume = sessiondata.coreplayer->GetVolume();
  volume += arg_int;
  if (volume > 100)
    volume = 100;
  else if (volume < 0)
    volume = 0;
  sessiondata.coreplayer->SetVolume(volume);
}

void CommandSet::vol_down(const std::string &args)
{
  int volume, arg_int;
  if (args.size() == 0U)
    arg_int = 5;
  else
    arg_int = atoi(args.c_str());
  volume = sessiondata.coreplayer->GetVolume();
  volume -= arg_int;
  if (volume > 100)
    volume = 100;
  else if (volume < 0)
    volume = 0;
  sessiondata.coreplayer->SetVolume(volume);
}

void CommandSet::pan_left(const std::string &args)
{
  int pan, arg_int;
  if (args.size() == 0U)
    arg_int = 5;
  else
    arg_int = atoi(args.c_str());
  pan = sessiondata.coreplayer->GetPan();
  pan -= arg_int;
  if (pan > 100)
    pan = 100;
  else if (pan < -100)
    pan = -100;
  sessiondata.coreplayer->SetPan(pan);
}

void CommandSet::pan_right(const std::string &args)
{
  int pan, arg_int;
  if (args.size() == 0U)
    arg_int = 5;
  else
    arg_int = atoi(args.c_str());
  pan = sessiondata.coreplayer->GetPan();
  pan += arg_int;
  if (pan > 100)
    pan = 100;
  else if (pan < -100)
    pan = -100;
  sessiondata.coreplayer->SetPan(pan);
}

void CommandSet::pan(const std::string &args)
{
  int arg_int;
  if (args.size() == 0U)
    arg_int = 0;
  else
    arg_int = atoi(args.c_str());
  sessiondata.coreplayer->SetPan(arg_int);
}

void CommandSet::faster(const std::string &args)
{
  float speed, arg_int;
  if (args.size() == 0U)
    arg_int = 20;
  else
    arg_int = atoi(args.c_str());
  speed = sessiondata.coreplayer->GetSpeed();
  speed += (double) arg_int / 100.0;
  sessiondata.coreplayer->SetSpeed(speed);
}

void CommandSet::slower(const std::string &args)
{
  float speed, arg_int;
  if (args.size() == 0U)
    arg_int = 20;
  else
    arg_int = atoi(args.c_str());
  speed = sessiondata.coreplayer->GetSpeed();
  speed -= (double) arg_int / 100.0;
  sessiondata.coreplayer->SetSpeed(speed);
}

void CommandSet::speed(const std::string &args)
{
  float arg_int;
  if (args.size() == 0U)
    arg_int = 100;
  else
    arg_int = atoi(args.c_str());
  sessiondata.coreplayer->SetSpeed((double) arg_int / 100.0);
}

void CommandSet::show_playlist(const std::string &args)
{
  sessiondata.changeMode(MODE_PLAYLIST);
}

void CommandSet::hide_playlist(const std::string &args)
{
  sessiondata.changeMode(MODE_PLAYER);
}

void CommandSet::playlist_up(const std::string &args)
{
  int arg_int;
  if (args.size() == 0U)
    arg_int = 1;
  else
    arg_int = atoi(args.c_str());

  if (sessiondata.playlistInterface->getSize() == 0)
    return;
  if (selectedSong < 0) // nothing selected -- use current
    selectedSong = sessiondata.playlist->GetCurrent() - 1;

  selectedSong -= arg_int;
  if (selectedSong < 0)
    selectedSong = 0;
  else if (selectedSong > sessiondata.playlistInterface->getSize())
    selectedSong = sessiondata.playlistInterface->getSize();
  sessiondata.display->setSelectedSong(selectedSong);
}

void CommandSet::playlist_down(const std::string &args)
{
  int arg_int;
  if (args.size() == 0U)
    arg_int = 1;
  else
    arg_int = atoi(args.c_str());

  if (sessiondata.playlistInterface->getSize() == 0)
    return;
  if (selectedSong < 0) // nothing selected -- use current
    selectedSong = sessiondata.playlist->GetCurrent() - 1;

  selectedSong += arg_int;
  if (selectedSong < 0)
    selectedSong = 0;
  else if (selectedSong > sessiondata.playlistInterface->getSize())
    selectedSong = sessiondata.playlistInterface->getSize();
  sessiondata.display->setSelectedSong(selectedSong);
}

void CommandSet::playlist_next_mode(const std::string &args)
{
  playlistMode = playlistmode_t((playlistMode + 1) % NUM_PLMODES);
  sessiondata.display->setPlaylistMode(playlistMode);
}

void CommandSet::playlist_prev_mode(const std::string &args)
{
  playlistMode = playlistmode_t((playlistMode + NUM_PLMODES - 1) % NUM_PLMODES);
  sessiondata.display->setPlaylistMode(playlistMode);
}

void CommandSet::playlist_select(const std::string &args)
{
  switch (playlistMode)
  {
  case PLMODE_PLAY:
    if (selectedSong >= 0
	&& selectedSong < sessiondata.playlistInterface->getSize())
    {
      sessiondata.playlist->Play(selectedSong+1);
    }
    break;
  case PLMODE_INSERT:
    sessiondata.changeMode(MODE_SELECT);
    break;
  case PLMODE_CLEAR:
    sessiondata.playlist->Clear();
    break;
  case PLMODE_DELETE:
    if (selectedSong >= 0
	&& selectedSong < sessiondata.playlistInterface->getSize())
    {
      sessiondata.playlist->Remove(selectedSong+1, selectedSong+1);
    }
    break;
  case NUM_PLMODES:
    break;
  }
}

void CommandSet::exit_selector(const std::string &args)
{
  sessiondata.changeMode(MODE_PLAYLIST);
}

void CommandSet::selector_up(const std::string &args)
{
  int arg_int;
  if (args.size() == 0U)
    arg_int = 1;
  else
    arg_int = atoi(args.c_str());

  selectedItemHistory.top() -= arg_int;
  if (selectedItemHistory.top() < 0)
    selectedItemHistory.top() = 0;
  else if (selectedItemHistory.top() >= sessiondata.selector->getSize())
    selectedItemHistory.top() = sessiondata.selector->getSize() - 1;

  sessiondata.display->setSelection(selectedItemHistory.top());
}

void CommandSet::selector_down(const std::string &args)
{
  int arg_int;
  if (args.size() == 0U)
    arg_int = 1;
  else
    arg_int = atoi(args.c_str());

  selectedItemHistory.top() += arg_int;
  if (selectedItemHistory.top() < 0)
    selectedItemHistory.top() = 0;
  else if (selectedItemHistory.top() >= sessiondata.selector->getSize())
    selectedItemHistory.top() = sessiondata.selector->getSize() - 1;

  sessiondata.display->setSelection(selectedItemHistory.top());
}

void CommandSet::selector_sublist(const std::string &args)
{
  if (sessiondata.selector->loadSublist(selectedItemHistory.top()))
  {
    selectedItemHistory.push(0);
    sessiondata.display->updateSelectionList();
    sessiondata.display->setSelection(0);
  }
}

void CommandSet::selector_backout(const std::string &args)
{
  if (sessiondata.selector->backout())
  {
    selectedItemHistory.pop();
    sessiondata.display->updateSelectionList();
    sessiondata.display->setSelection(selectedItemHistory.top());
  }
  else
    sessiondata.changeMode(MODE_PLAYLIST);
}

void CommandSet::selector_select(const std::string &args)
{
  sessiondata.selector->addToPlaylist(selectedItemHistory.top(),
				      sessiondata.playlist,
				      selectedSong < 0
				      ? sessiondata.playlistInterface->getSize()
				      : selectedSong);
}

void CommandSet::show_menu(const std::string &args)
{
  sessiondata.changeMode(MODE_MENU);
}

void CommandSet::exit_menu(const std::string &args)
{
  sessiondata.changeMode(MODE_PLAYER);
}

void CommandSet::menu_next_pref(const std::string &args)
{
  int arg_int;
  if (args.size() == 0U)
    arg_int = 1;
  else
    arg_int = atoi(args.c_str());

  selectedPref += arg_int;
  if (selectedPref < 0)
    selectedPref = 0;
  else if (selectedPref >= sessiondata.preferences->getSize())
    selectedPref = sessiondata.preferences->getSize() - 1;

  sessiondata.display->setPreference(selectedPref);
}

void CommandSet::menu_prev_pref(const std::string &args)
{
  int arg_int;
  if (args.size() == 0U)
    arg_int = 1;
  else
    arg_int = atoi(args.c_str());

  selectedPref -= arg_int;
  if (selectedPref < 0)
    selectedPref = 0;
  else if (selectedPref >= sessiondata.preferences->getSize())
    selectedPref = sessiondata.preferences->getSize() - 1;

  sessiondata.display->setPreference(selectedPref);
}

void CommandSet::menu_next_value(const std::string &args)
{
  sessiondata.preferences->getItem(selectedPref).nextValue();
  sessiondata.display->updatePreference(selectedPref);
}

void CommandSet::menu_prev_value(const std::string &args)
{
  sessiondata.preferences->getItem(selectedPref).prevValue();
  sessiondata.display->updatePreference(selectedPref);
}


CommandSet::CommandSet(SessionData &_sessiondata)
  : ModeSubscriber(_sessiondata),
    curr_mode(&callbacks[sessiondata.getMode()]),
    oldspeed(1.0),
    selectedSong(-1),
    playlistMode(PLMODE_PLAY),
    selectedPref(0)
{
  if (!is_initialized)
    init();
  selectedItemHistory.push(0);
}

// initialize the callback table
void CommandSet::init (void)
{
  if (!is_initialized)
  {
    // these commands are active in all modes -- basically all core
    // player commands
    callbacks[MODE_ALL]["play"] = &CommandSet::play;
    callbacks[MODE_ALL]["stop"] = &CommandSet::stop;
    callbacks[MODE_ALL]["pause"] = &CommandSet::pause;
    callbacks[MODE_ALL]["playpause"] = &CommandSet::playpause;
    callbacks[MODE_ALL]["next"] = &CommandSet::next;
    callbacks[MODE_ALL]["prev"] = &CommandSet::prev;
    callbacks[MODE_ALL]["fwd"] = &CommandSet::fwd;
    callbacks[MODE_ALL]["bwd"] = &CommandSet::bwd;
    callbacks[MODE_ALL]["vol_up"] = &CommandSet::vol_up;
    callbacks[MODE_ALL]["vol_down"] = &CommandSet::vol_down;
    callbacks[MODE_ALL]["pan_left"] = &CommandSet::pan_left;
    callbacks[MODE_ALL]["pan_right"] = &CommandSet::pan_right;
    callbacks[MODE_ALL]["pan"] = &CommandSet::pan;
    callbacks[MODE_ALL]["faster"] = &CommandSet::faster;
    callbacks[MODE_ALL]["slower"] = &CommandSet::slower;
    callbacks[MODE_ALL]["speed"] = &CommandSet::speed;

    // these commands are active only in the default player mode
    callbacks[MODE_PLAYER]["left"] = &CommandSet::pan_left;
    callbacks[MODE_PLAYER]["right"] = &CommandSet::pan_right;
    callbacks[MODE_PLAYER]["up"] = &CommandSet::faster;
    callbacks[MODE_PLAYER]["down"] = &CommandSet::slower;

    // these commands are used for manipulating the playlist
    callbacks[MODE_PLAYER]["playlist"] = &CommandSet::show_playlist;
    callbacks[MODE_PLAYLIST]["playlist"] = &CommandSet::hide_playlist;
    callbacks[MODE_PLAYLIST]["exit"] = &CommandSet::hide_playlist;
    callbacks[MODE_PLAYLIST]["up"] = &CommandSet::playlist_up;
    callbacks[MODE_PLAYLIST]["down"] = &CommandSet::playlist_down;
    callbacks[MODE_PLAYLIST]["right"] = &CommandSet::playlist_next_mode;
    callbacks[MODE_PLAYLIST]["left"] = &CommandSet::playlist_prev_mode;
    callbacks[MODE_PLAYLIST]["select"] = &CommandSet::playlist_select;

    // these commands are used for the selection screen
    callbacks[MODE_SELECT]["playlist"] = &CommandSet::exit_selector;
    callbacks[MODE_SELECT]["exit"] = &CommandSet::exit_selector;
    callbacks[MODE_SELECT]["up"] = &CommandSet::selector_up;
    callbacks[MODE_SELECT]["down"] = &CommandSet::selector_down;
    callbacks[MODE_SELECT]["right"] = &CommandSet::selector_sublist;
    callbacks[MODE_SELECT]["left"] = &CommandSet::selector_backout;
    callbacks[MODE_SELECT]["select"] = &CommandSet::selector_select;

    // these commands are used for the preferences menu
    callbacks[MODE_PLAYER]["menu"] = &CommandSet::show_menu;
    callbacks[MODE_MENU]["menu"] = &CommandSet::exit_menu;
    callbacks[MODE_MENU]["exit"] = &CommandSet::exit_menu;
    callbacks[MODE_MENU]["down"] = &CommandSet::menu_next_pref;
    callbacks[MODE_MENU]["up"] = &CommandSet::menu_prev_pref;
    callbacks[MODE_MENU]["right"] = &CommandSet::menu_next_value;
    callbacks[MODE_MENU]["left"] = &CommandSet::menu_prev_value;


    is_initialized = true;
  }
}

void CommandSet::call(const std::string &command, const std::string &args)
{
  if (all_modes.find(command) != all_modes.end())
    (this->*all_modes[command])(args);
  else if (curr_mode->find(command) != curr_mode->end())
    (this->*(*curr_mode)[command])(args);
}

